package ch.zhaw.pm2.multichat.client;

import ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State;
import ch.zhaw.pm2.multichat.protocol.ChatProtocolException;
import ch.zhaw.pm2.multichat.protocol.ConnectionHandler;
import ch.zhaw.pm2.multichat.protocol.NetworkHandler;
import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.TextArea;
import javafx.scene.control.TextField;
import javafx.scene.layout.Pane;
import javafx.stage.WindowEvent;

import java.io.IOException;

import static ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State.*;

public class ChatWindowController {
    private ClientConnectionHandler connectionHandler;
    private ClientMessageList messages;

    private final WindowCloseHandler windowCloseHandler = new WindowCloseHandler();

    @FXML private Pane rootPane;
    @FXML private TextField serverAddressField;
    @FXML private TextField serverPortField;
    @FXML private TextField userNameField;
    @FXML private TextField messageField;
    @FXML private TextArea messageArea;
    @FXML private Button connectButton;
    @FXML private Button sendButton;
    @FXML private TextField filterValue;


    @FXML
    public void initialize() {
        serverAddressField.setText(NetworkHandler.DEFAULT_ADDRESS.getCanonicalHostName());
        serverPortField.setText(String.valueOf(NetworkHandler.DEFAULT_PORT));
    }

    public void setMessages(ClientMessageList messages) {
        this.messages = messages;
        messageListener();
    }

    public void setConnectionHandler(ClientConnectionHandler connectionHandler){
        this.connectionHandler = connectionHandler;
    }

    private void applicationClose() {
        disconnect();
    }

    @FXML
    private void toggleConnection () {
        if (connectionHandler == null || connectionHandler.getStateProperty().get() != CONNECTED) {
            connect();
        } else {
            disconnect();
        }
    }

    private void connect() {
        try {
            messages.clear(); // clear message list
            startConnectionHandler();
            connectionHandler.connect();
        } catch(ChatProtocolException | IOException e) {
            addError(e.getMessage());
        }
    }

    private void disconnect() {
        if (connectionHandler == null) {
            addError("No connection handler");
            return;
        }
        try {
            connectionHandler.disconnect();
        } catch (ChatProtocolException e) {
            addError(e.getMessage());
        }
    }

    @FXML
    private void message() {
        String messageString = messageField.getText().strip();
        try {
            if (connectionHandler == null) {
                addError("No connection handler");
            } else if (!connectionHandler.message(messageString)) {
                addError("Not a valid message format.");
            } else {
                messageField.clear();
            }
        } catch (ChatProtocolException e) {
            addError(e.getMessage());
        }
    }

    @FXML
    private void applyFilter( ) {
    	this.redrawMessageList();
    }

    private void startConnectionHandler() throws IOException {
        String userName = userNameField.getText();
        if(!userName.contains(" ")) {
            String serverAddress = serverAddressField.getText();
            int serverPort = Integer.parseInt(serverPortField.getText());
            connectionHandler.initialize(serverAddress, serverPort, userName);
            new Thread(connectionHandler).start();

            //register Listener
            startListener();

            // register window close handler
            rootPane.getScene().getWindow().addEventHandler(WindowEvent.WINDOW_CLOSE_REQUEST, windowCloseHandler);
        } else {
            addError("It is not allowed to have spaces in username!");
        }
    }

    public void stateChanged(State newState) {
        // update UI (need to be run in UI thread: see Platform.runLater())
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                connectButton.setText((newState == CONNECTED || newState == CONFIRM_DISCONNECT) ? "Disconnect" : "Connect");
            }
        });
        if(newState == DISCONNECTED){
            connectionHandler.stopReceiving();
        }
    }

    public void setUserName(String userName) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                userNameField.setText(userName);
            }
        });
    }

    public void setServerAddress(String serverAddress) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                serverAddressField.setText(serverAddress);
            }
        });
    }

    public void setServerPort(int serverPort) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                serverPortField.setText(Integer.toString(serverPort));
            }
        });
    }

    public void addError(String message) {
        messages.addMessage(new Message(Message.MessageType.ERROR, null, null, message));
    }

    private void redrawMessageList() {
        this.messageArea.clear();
        Platform.runLater(() -> this.messageArea.setText(messages.getFilteredMessages(filterValue.getText().strip())));
    }

    class WindowCloseHandler implements EventHandler<WindowEvent> {
        public void handle(WindowEvent event) {
            applicationClose();
        }

    }

    public void startListener() {
        connectionHandler.getStateProperty().addListener(new ChangeListener<State>() {
            @Override
            public void changed(ObservableValue<? extends State> observable, State oldValue, State newValue) {
                stateChanged(newValue);
            }
        });

        connectionHandler.getUserNameProperty().addListener(new ChangeListener<String>() {
            @Override
            public void changed(ObservableValue<? extends String> observable, String oldValue, String newValue) {
                setUserName(newValue);
            }
        });

        connectionHandler.getServerAddressProperty().addListener(new ChangeListener<String>() {
            @Override
            public void changed(ObservableValue<? extends String> observable, String oldValue, String newValue) {
                setServerAddress(newValue);
            }
        });

        connectionHandler.getServerPortProperty().addListener(new ChangeListener<Number>() {
            @Override
            public void changed(ObservableValue<? extends Number> observable, Number oldValue, Number newValue) {
                setServerPort(newValue.intValue());
            }
        });
    }

    private void messageListener() {
        messages.getChangedProperty().addListener(new ChangeListener<Boolean>() {
            @Override
            public void changed(ObservableValue<? extends Boolean> observable, Boolean oldValue, Boolean newValue) {
                redrawMessageList();
            }
        });
    }

}
